import abc
import binascii
import logging
from base64 import urlsafe_b64encode, urlsafe_b64decode
from collections import OrderedDict
from collections.abc import Iterable

from cryptography.fernet import Fernet, InvalidToken

from slashdb.exceptions import (
    EmptySecret,
    InvalidSecret,
    InvalidCipher,
    InvalidSignature,
    DecryptionFailed,
    EncryptionError,
)
from slashdb.constants import ConnectStatus
from slashdb.validation import ValidationErrorType

log = logging.getLogger('slashdb')

MODULE_MAKERS = {}
DATA_PROXIES = {}
DB_CONFIGS = {}


class IModuleMaker(object):  # pragma: no cover
    """Database model interface."""

    def __init__(self, db_id, db_config, ini_settings):
        """
        :param db_id - a string representing the data sources unique identifier
        :param db_config - a dict-like object, representing the data sources configuration
        :param ini_settings - a dictionary representing the parsed configuration *.ini file
        """
        self.db_id = db_id
        self.db_config = db_config
        self.ini_settings = ini_settings

    def create(self):
        """Returns a new python module with classes that have implemented IModel interface.

        Basically
        imp.new_module(self.db_id)
        new_class = type(str(class_name), (ModelClass,), {'_field_names': field_names})
        setattr(new_module, new_class.__name__, new_class)
        return new_module

        In above short example
        class_name - name of a new class
        ModelClass - class template that implements IModel from plugin package
        {'_field_names': field_names} - class attributes of new class

        """
        raise NotImplementedError


class DictMixin:
    """Part of DictMixin Backported from UserDict

    This class was removed in python3. In order to migrate only actually used
    methods were copied
    """
    def values(self):
        for _, v in self.items():
            yield v

    def update(self, other=None, **kwargs):
        # Make progressively weaker assumptions about "other"
        if other is None:
            pass
        elif hasattr(other, 'items'):  # iteritems saves memory and lookups
            for k, v in other.items():
                self[k] = v
        elif hasattr(other, 'keys'):
            for k in other.keys():
                self[k] = other[k]
        else:
            for k, v in other:
                self[k] = v
        if kwargs:
            self.update(kwargs)

    def __len__(self):
        return len(self.keys())


class IModel(DictMixin):  # pragma: no cover
    """Basic resource/model interface.

    Reflected models are expected to implement these methods for proper rendering.
    Some methods are required to make this object behave like a dictionary.
    """

    def __getitem__(self, param):
        raise NotImplementedError

    def __setitem__(self, key, value):
        raise NotImplementedError

    def __len__(self):
        raise NotImplementedError

    def __iter__(self):
        raise NotImplementedError

    def keys(self):
        raise NotImplementedError

    def items(self):
        raise NotImplementedError

    def get_href(self, field_name=None, build_escaped_url=None, extension=None):
        """Returns href without domain to resource or field or relation of the resource."""
        raise NotImplementedError

    def get_href_segments(self, field_name=None, build_escaped_url=None):
        """Returns href without domain to resource or field or relation of the resource as list of url segments."""
        raise NotImplementedError

    @classmethod
    def get_arranged_props(cls):
        """List of strings with names of attributes/keys in preferred order (columns and relations)."""
        raise NotImplementedError

    @classmethod
    def get_column_names(cls):
        """List of string with names of attributes/keys that refer to columns."""
        raise NotImplementedError

    @classmethod
    def get_relation_names(cls):
        """List of strings with names of attributes/keys that refer to relationships in object."""
        raise NotImplementedError

    @classmethod
    def get_related_class(cls, relation_name):
        """Returns a related class by given name of the relation (object attribute)."""
        raise NotImplementedError

    @classmethod
    def get_python_type(cls, column_name):
        """Returns a correct python type for column data."""
        raise NotImplementedError

    @classmethod
    def get_python_value(cls, column_name, text_value):
        """Converts text to proper python type value."""
        raise NotImplementedError

    @classmethod
    def get_attr_type(cls, attr_name):
        """Returns proper slashdb.dbproxies.*.models.AttrType for queried attribute/key."""
        raise NotImplementedError

    def __repr__(self):
        return '<{}>'.format(type(self).__name__)

    def check_ambiguous_columns(self):
        raise NotImplementedError


class IDataProxy(object):  # pragma: no cover
    """Data proxy interface.

    Objects of that class implement functionality of
    getting, streaming, updating, creating and deleting source data.

    Attributes:
        db_types (list): list of strings representing database types supported by this data proxy.
            This also works as an identifier in license.

        is_single_resource (bool): True if present output as single resource
            For instance when PK filtering is applied.

        resource_context: slashdb.datastructures.ResourceMeta object with details for the renderers
            ResourceMeta(
                resource_class=TheClass,        # class of IModel
                scalars=['field_a', 'field_b']  # list of strings with names of fields for vector or array
            )

    """

    db_types = None
    is_single_resource = None
    resource_context = None

    class TempProcessRegistry:
        """Mocks process registry to provide ini settings.

        Needed by dataproxies in method _make_temp_db_config, the database config
        needs sdb_settings to decrypt passwords.
        """

        def __init__(self, ini_settings):
            self.ini_settings = ini_settings
            self.sdb_settings = ini_settings['sdb_settings']

    def __init__(self, context, request):
        """
        Args:
            context: instance of slashdb.roots.DataDiscovery of slashdb.roots.Query.
                Context provides
                    context.url_segments - an iterable of information caring parts of the url i.e. ('SomeDB', 'MyTable', 'MyColumn')
                    context.db_id - database unique ID
                    context.db_config - a dict-like object, representing the data sources configuration
                    context.query_id  - it's None if used DataDiscovery, way check if SQL Pass-thru
                    context.query_config  - it's None if used DataDiscovery
                    context.authenticated_user_id - ID of the logged-in user
                    context.authenticated_user_config - a dict-like object, representing the users configuration
                    context.settings  - settings prepared on start. Consider as deprecated.
                        This should not be used as trying to remove it from application.
                    context.response_format  - requested output format based on extension and mime type.
                        This should be bound to available renderers formats.
                        Right now this is used as key to pick renderer or parser from
                        slashdb.formats.RENDERERS or slashdb.formats.PARSERS.

            request (pyramid.request.Request): pyramid request which is source of URL params, and ini settings

        """
        self.context = context
        self.request = request

    @staticmethod
    def check_database_connection(ini_settings, server_info):
        """Method called to test if connection is possible.

        :param ini_settings: parsed ini settings
        :param server_info: Dictionary that contains server information for
        example
        {
            "database": "Chinook",
            "db_type": "mssql",
            "file_path": "",  # for sqlite
            "hostname": "192.168.1.12",
            "password": "p4ssw0rdD",
            "port": "",
            "schema": "",
            "user"": "chinook"
        }

        """
        raise NotImplementedError

    @staticmethod
    def check_database_connection_from_connstr(ini_settings, connstr, schema, engine_kwargs = None):
        raise NotImplementedError

    def get(self, stream, flatten):
        """Method called by renderers to acquire data.

        Streaming data is optional allows to query data in chunks and render
        sequentially by renderer. That would reduce memory usage.

        Args:
            stream (bool): if True an iterator will be returned instead of list
            flatten (bool): if True the returned data will be flattened
        """
        raise NotImplementedError

    def create(self, data):
        """Method that will be called to save new data into the resource.

        It is called by slashdb.data_discovery_view.post_view. Together with information from context
        it should be sufficient to commit new data to database.

        Args:
            data (dict): with data for new resource

        """
        raise NotImplementedError

    def update(self, update_data):
        """Method that will be called to update data in the resource.

        It is called by slashdb.data_discovery_view.put_view. Together with information from context
        it should be sufficient to commit new data to database.

        Args:
            update_data (dict): with data for new resource

        """
        raise NotImplementedError

    def delete(self):
        """Method that will be called to delete data from the resource."""
        raise NotImplementedError

    def close(self):
        """Method to be called when we are done with using this resource.

        Useful when there's a need to do nice cleanup on the db session or file.
        Pyramid calls app_iter.close() which should call data_proxy.close().

        """

    def response_postprocessing(self, response):
        """Method for modifying freshly created response.

        Allows data proxy to make some changes in response like add a header.

        Args:
            response (pyramid.response.Response): created response object
                with defined body or app_iter, content type and charset


        Returns:
             None

        """
        pass

    @classmethod
    def get_engine_kwargs(
            cls, db_type:str, db_id:str, ini_settings, is_insert_safe:bool = False, is_streaming: bool = False
            ) -> dict:
        """Creates options that are passed to SQLAlchemy create_engine."""
        return {}

class CleanBase(object):
    """A simple class, adding field change detection.

    Setting some attribute for the first time does not make object dirty.
    Object becomes dirty when attribute is updated.

    """

    _keys = ()
    _clean = True
    # special value to distinguish between a situation where a
    # field has been explicitly set to None
    # and when getattr returns the default None
    # i.e. field was not set at all
    _not_yet_set = 'NotYetSet'

    def __setattr__(self, name, value, nys=_not_yet_set):
        old_value = getattr(self, name, nys)
        if old_value != nys and name in self._keys:
            if value != old_value:
                self._clean = False
            else:
                return
        super(CleanBase, self).__setattr__(name, value)

    def __delattr__(self, name):
        if name in self._keys:
            self._clean = False
        super(CleanBase, self).__delattr__(name)

    @property
    def is_dirty(self):
        """Check if fields declared in _keys were modified."""
        return not self._clean

    def set_clean(self):
        """Sets instance state to clean."""
        self._clean = True


class ValidationMixin(object):
    def __init__(self, *args, **kwargs):
        # validation attributes
        self.errors = []
        self.conflict = False

    def add_error(self, msg, field):
        self.errors.append({
            'field': field,
            'type': ValidationErrorType.invalid,
            'description': msg
        })

    def validate(self, **kwargs):
        """Validates config data.

        :param kwargs: a dictionary containing any additional info that the validate method might want to use

        """
        self.reset_validation()

    def reset_validation(self):
        """Clears validation result."""
        self.errors = []
        self.conflict = False

    def _check_config_id_mismatch(self, dict_key):
        if dict_key is None:
            return

        if dict_key != self.config_id:
            msg = '"{}" did not match configuration id "{}"'.format(
                self.config_id, dict_key
            )
            self.add_error(msg, self._ID_KEY)

    def _validate_creator(self):
        if self.creator is None:
            self.creator = 'admin'


class IDBConfig(CleanBase, ValidationMixin, metaclass=abc.ABCMeta):

    """Abstract class for SlashDB configuration objects."""

    _log_type = 'database'
    _keys = [
        'db_id',
        'desc', 'autoload', 'autoconnect', 'cache_dbmodel',
        'db_type', 'db_dialect', 'db_driver',
        'db_host', 'db_port', 'db_name', 'db_schema', 'db_encoding',
        'connection',
        'autoload_user',
        'alternate_key', 'excluded_columns', 'foreign_keys',
        'creator', 'read', 'write', 'execute',
        'connect_status'
    ]

    def __init__(self, process_registry,
                 db_id,
                 desc, autoload, autoconnect,
                 db_type, db_dialect, db_driver,
                 db_host, db_port, db_name, db_schema, db_encoding,
                 connection,
                 autoload_user,
                 alternate_key, excluded_columns, foreign_keys,
                 creator, read, write, execute,
                 connect_status=None, cache_dbmodel=None):
        """Sets up DBConfig object.

        Validation can be skipped with validate=False if you're sure that attribute values are correct.

        :param db_id: model, database id
        :param db_type: type like sqlite, mssql etc.
        :param db_dialect: sqlalchemy database dialect
        like mysql for db_type mariadb
        :param db_driver: DBAPI name to be used for connection
        :param db_host: hostname or IP address of database server
        :param connection: connection string without db_type
        :param db_encoding: database charset encoding
        :param desc: more detail description of the database
        :param autoload_user: database user and password
        :param autoload: if True then SlashDB will try to reflect database
        :param autoconnect: if True then SlashDB will connect database automatically on start
        :param cache_dbmodel: if True then SlashDB will save and load the database metadata from a pickle
        :param creator: id of user that added this database config
        :param read: users allowed to read this database config
        :param write: users allowed to read and change this database config
        :param execute: users allowed to connect/disconnect this database on demand
        :param db_schema: database schema
        :param alternate_key: list of columns to use as composite primary key (to be used with autoload=True)
        :param excluded_columns: list of columns to exclude from table (to be used with autoload=True)
        :param connect_status: state offline/connecting/connected/failed
        :param foreign_keys: a dict of user defined relations
        :param process_registry: ProcessRegistry object used for
        synchronization of db reflect status

        """

        super(IDBConfig, self).__init__()

        self.process_registry = process_registry
        self.ini_settings = self.process_registry.ini_settings
        self.sdb_settings = self.ini_settings['sdb_settings']

        self.db_id = db_id
        self.db_type = db_type
        self.db_dialect = db_dialect
        self.db_driver = db_driver
        self.db_host = db_host
        self.db_port = db_port
        self.db_name = db_name
        self.connection = connection
        self.db_encoding = db_encoding
        self.desc = desc
        self.autoload_user = autoload_user or {}
        self.autoload = autoload
        self.autoconnect = autoconnect
        self.cache_dbmodel = cache_dbmodel
        self.creator = creator
        self.read = read
        self.write = write
        self.execute = execute
        self.db_schema = db_schema
        self.alternate_key = alternate_key
        self.excluded_columns = excluded_columns
        self.foreign_keys = foreign_keys if foreign_keys else {}
        self.connect_status = connect_status or ConnectStatus.OFF
        self.connect_message = ''
        self.reflect_status = {}

        self.errors = []
        self.warnings = []

        self.secret = self.sdb_settings['secret']
        dbpass = self.autoload_user.get('dbpass')
        if self.secret and dbpass is not None:
            sdbc = SlashDBCrypto(self.secret)

            try:
                # attempt to decrypt password with current secret
                # No error means everything is OK and decryption will be
                # possible.
                sdbc.decrypt(dbpass)
            except InvalidSignature:
                # when stored dbpass is not encrypted with the current
                # secret then use plaintext
                log.warning(
                    f'Encryption key does not match database password '
                    f'cipher. Autoload user for database config {self.db_id}.'
                )
            except InvalidCipher:
                log.info(f'Database password is not a valid cipher. '
                         f'Autoload user for database config {self.db_id}.'
                         f'Password will be encrypted with current secret.')
                # when stored dbpass is not recognized as cipher at all
                # then encrypt it
                self.autoload_user['dbpass'] = sdbc.encrypt(dbpass)

    def __unicode__(self):
        return "<{}>".format(type(self).__name__)

    def __repr__(self):
        return "<{}>".format(type(self).__name__)

    @property
    def config_id(self):
        return self.db_id

    @property
    def simple_dict(self):
        return {key: getattr(self, key) for key in self._keys}

    def get_storage_dict(self, skip=None):
        skip = skip or []
        storage_dict = self.simple_dict
        for key in skip:
            storage_dict.pop(key, None)
        return storage_dict

    @property
    def is_connected(self):
        return self.connect_status == ConnectStatus.ON

    def _validate_autoload_user(self):
        """
        Makes sure autoload_user is dictionary with keys dbuser and dbpass.
        """
        if self.autoload_user in [None, {}]:
            self.autoload_user = {'dbuser': None, 'dbpass': None}

        self.autoload_user['dbuser'] = self.autoload_user.get('dbuser')
        self.autoload_user['dbpass'] = self.autoload_user.get('dbpass')

    def _validate_cache_dbmodel(self):
        """Makes sure cache_dbmodel is a boolean"""
        self.cache_dbmodel = bool(self.cache_dbmodel)

    @abc.abstractmethod
    def validate(self):
        """Validates config data."""
        return super(IDBConfig, self).validate()

    def add_to_process_registry(self):
        self.process_registry.add_config(self)

    def get_db_credentials(self):
        """Returns dbuser and decrypted dbpass from autoload user mapping."""

        dbuser = self.autoload_user.get('dbuser')
        dbpass = self.autoload_user.get('dbpass')

        if self.secret and dbpass:
            sdbc = SlashDBCrypto(self.secret)
            try:
                dbpass = sdbc.decrypt(dbpass)
            except EncryptionError:
                # in case problem with decrypting use plaintext
                log.warning(f'Database password could not be decrypted. '
                            f'Autoload user for database config {self.db_id}.')

        return dbuser, dbpass


def parse_as_list(permissions):
    """Converts given permissions to a list while removing duplicates.

    :param permissions: Iterable or string containing the permissions
    :returns tuple(list, bool)

    """
    parts = []
    is_invalid_type = False

    if not permissions:
        return parts, is_invalid_type

    if isinstance(permissions, str):
        parts = permissions.split(',')
    elif isinstance(permissions, Iterable):
        parts = (i for i in permissions if isinstance(i, str))
    else:
        is_invalid_type = True

    def clean_parts():
        for part in parts:
            part = part.strip()
            if part != '':
                yield part

    # an ad hoc 'ordered set' based on OrderedDict().viewkeys()
    parts = list(OrderedDict(
        (k, None) for k in clean_parts()
    ).keys())
    return parts, is_invalid_type


class SlashDBCrypto(Fernet):
    """Enhanced version of Fernet which works with string values."""
    def __init__(self, secret):
        # check secret
        if not secret:
            raise EmptySecret

        if isinstance(secret, str):
            secret = secret.encode()

        if len(secret) == 32:
            # assuming the secret is alphanumeric so it's converted to base64
            secret = urlsafe_b64encode(secret)
        else:
            # assuming the secret is base64 but checking
            try:
                urlsafe_b64decode(secret)
            except binascii.Error:
                raise InvalidSecret

        try:
            super().__init__(secret)
        except ValueError:
            raise InvalidSecret

    def encrypt(self, text):
        """Encrypts the text and return cipher as str.

        Fernet returns cipher as base64, ASCII byte string.
        The cipher is decoded with ASCII to return str type.
        """
        if isinstance(text, str):
            text = text.encode()
        cipher = super().encrypt(text)

        return cipher.decode('ascii')

    def decrypt(self, cipher):
        """Returns decrypted cipher as str.

        Cipher must be base64, either ASCII byte string or str type.
        """

        # encoding str because Fernet needs bytes
        if isinstance(cipher, str):
            try:
                cipher = cipher.encode('ascii')
            except UnicodeEncodeError:
                raise InvalidCipher

        # check cipher
        try:
            timestamp, data = Fernet._get_unverified_token_data(cipher)
        except InvalidToken:
            raise InvalidCipher

        # check cipher matches secret
        try:
            self._verify_signature(data)
        except InvalidToken:
            raise InvalidSignature

        # decrypt
        try:
            text = super().decrypt(cipher)
        except InvalidToken:
            raise DecryptionFailed

        text = text.decode()
        return text
